#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions and class related with maths
"""

from __future__ import print_function, division, absolute_import

import math
import struct

MAX_INT = 2 ** (struct.Struct('i').size * 8 - 1) - 1


class BaseVector(object):
    pass


class Vector2D(object):
    def __init__(self, x=1.0, y=1.0):
        self.x = None
        self.y = None

        if type(x) == list or type(x) == tuple:
            self.x = x[0]
            self.y = x[1]

        if type(x) == float or type(x) == int:
            self.x = x
            self.y = y

        self.magnitude = None

    def _add(self, value):
        if type(value) == float or type(value) == int:
            return Vector2D(self.x + value, self.y + value)

        if type(self) == type(value):
            return Vector2D(value.x + self.x, value.y + self.y)

        if type(value) == list:
            return Vector2D(self.x + value[0], self.y + value[1])

    def _sub(self, value):
        if type(value) == float or type(value) == int:
            return Vector2D(self.x - value, self.y - value)

        if type(self) == type(value):
            return Vector2D(self.x - value.x, self.y - value.y)

        if type(value) == list:
            return Vector2D(self.x - value[0], self.y - value[1])

    def _rsub(self, value):
        if type(value) == float or type(value) == int:
            return Vector2D(value - self.x, value - self.y - value)

        if type(self) == type(value):
            return Vector2D(value.x - self.x, value.y - self.y)

        if type(value) == list:
            return Vector2D(value[0] - self.x, value[1] - self.y)

    def _mult(self, value):
        if type(value) == float or type(value) == int:
            return Vector2D(self.x * value, self.y * value)

        if type(self) == type(value):
            return Vector2D(value.x * self.x, value.y * self.y)

        if type(value) == list:
            return Vector2D(self.x * value[0], self.y * value[1])

    def _divide(self, value):
        if type(value) == float or type(value) == int:
            return Vector2D(self.x / value, self.y / value)

        if type(self) == type(value):
            return Vector2D(value.x / self.x, value.y / self.y)

        if type(value) == list:
            return Vector2D(self.x / value[0], self.y / value[1])

    def __add__(self, value):
        return self._add(value)

    def __radd__(self, value):
        return self._add(value)

    def __sub__(self, value):
        return self._sub(value)

    def __rsub__(self, value):
        return self._sub(value)

    def __mul__(self, value):
        return self._mult(value)

    def __rmul__(self, value):
        return self._mult(value)

    def __call__(self):
        return [self.x, self.y]

    def __div__(self, value):
        return self._divide(value)

    def _reset_data(self):
        self.magnitude = None

    def normalize(self, in_place=False):
        if not self.magnitude:
            self.magnitude()

        vector = self._divide(self.magnitude)

        if in_place:
            self.x = vector.x
            self.y = vector.y
            self._reset_data()

        if not in_place:
            return vector

    def get_vector(self):
        return [self.x, self.y]

    def get_magnitude(self):
        self.magnitude = math.sqrt((self.x * self.x) + (self.y * self.y))
        return self.magnitude

    def get_distance(self, x=0.0, y=0.0):
        other = Vector2D(x, y)

        offset = self - other

        return offset.get_magnitude()


class Vector(object):
    def __init__(self, x=1.0, y=1.0, z=1.0):

        self.x = None
        self.y = None
        self.z = None

        x_test = x

        if type(x_test) == list or type(x_test) == tuple:
            self.x = x[0]
            self.y = x[1]
            self.z = x[2]

        if type(x_test) == float or type(x_test) == int:
            self.x = x
            self.y = y
            self.z = z

        if isinstance(x_test, Vector):
            self.x = x_test.x
            self.y = x_test.y
            self.z = x_test.z

    def _add(self, value):
        if type(value) == float or type(value) == int:
            return Vector(self.x + value, self.y + value, self.z + value)

        if type(self) == type(value):
            return Vector(value.x + self.x, value.y + self.y, value.z + self.z)

        if type(value) == list:
            return Vector(self.x + value[0], self.y + value[1], self.z + value[2])

    def _sub(self, value):
        if type(value) == float or type(value) == int:
            return Vector(self.x - value, self.y - value, self.z - value)

        if type(self) == type(value):
            return Vector(self.x - value.x, self.y - value.y, self.z - value.z)

        if type(value) == list:
            return Vector(self.x - value[0], self.y - value[1], self.z - value[2])

    def _rsub(self, value):
        if type(value) == float or type(value) == int:
            return Vector(value - self.x, value - self.y - value, value - self.z)

        if type(self) == type(value):
            return Vector(value.x - self.x, value.y - self.y, value.z - self.z)

        if type(value) == list:
            return Vector(value[0] - self.x, value[1] - self.y, value[2] - self.z)

    def _mult(self, value):
        if type(value) == float or type(value) == int:
            return Vector(self.x * value, self.y * value, self.z * value)

        if type(self) == type(value):
            return Vector(value.x * self.x, value.y * self.y, value.z * self.z)

        if type(value) == list:
            return Vector(self.x * value[0], self.y * value[1], self.z * value[2])

    def __add__(self, value):
        return self._add(value)

    def __radd__(self, value):
        return self._add(value)

    def __sub__(self, value):
        return self._sub(value)

    def __rsub__(self, value):
        return self._sub(value)

    def __mul__(self, value):
        return self._mult(value)

    def __rmul__(self, value):
        return self._mult(value)

    def __call__(self):
        return [self.x, self.y, self.z]

    def get_vector(self):
        return [self.x, self.y, self.z]

    def list(self):
        return self.get_vector()


class BoundingBox(object):
    """
    Util class to work with bounding box
    """

    def __init__(self, bottom_corner_list=None, top_corner_list=None):
        """
        Constructor
        :param bottom_corner_list: list<float, float, float>, vector of bounding box bottom corner
        :param top_corner_list: list<float, float, float>, vector of bounding box top corner
        """

        self._create_bounding_box(bottom_corner_list, top_corner_list)

    def _create_bounding_box(self, bottom_corner_list, top_corner_list):
        """
        Initializes bounding box
        :param bottom_corner_list: list<float, float, float>, vector of bounding box bottom corner
        :param top_corner_list: list<float, float, float>, vector of bounding box top corner
        """

        self.min_vector = [bottom_corner_list[0], bottom_corner_list[1], bottom_corner_list[2]]
        self.max_vector = [top_corner_list[0], top_corner_list[1], top_corner_list[2]]
        self.opposite_min_vector = [top_corner_list[0], bottom_corner_list[1], top_corner_list[2]]
        self.opposite_max_vector = [bottom_corner_list[0], top_corner_list[1], bottom_corner_list[2]]

    def get_center(self):
        """
        Returns the center of the bounding box in a list
        :return: list<float, float, float>
        """

        return get_mid_point(self.min_vector, self.max_vector)

    def get_ymax_center(self):
        """
        Returns the top center of the bounding box in a list
        :return: list<float, float, float>
        """

        return get_mid_point(self.max_vector, self.opposite_max_vector)

    def get_ymin_center(self):
        """
        Returns the bottom center of the bounding box in a list
        :return: list<float, float, float>
        """

        return get_mid_point(self.min_vector, self.opposite_min_vector)

    def get_size(self):
        """
        Returns the size of the bounding box
        :return: float
        """

        return get_distance_between_vectors(self.min_vector, self.max_vector)


def is_equal(x, y, tolerance=0.000001):
    """
    Checks if 2 float values are equal withing a given tolerance
    :param x: float, first float value to compare
    :param y: float, second float value to compare
    :param tolerance: float, comparison tolerance
    :return: bool
    """

    return abs(x - y) < tolerance


def lerp(start, end, alpha):
    """
    Computes a linear interpolation between two values
    :param start: start value to interpolate from
    :param end:  end value to interpolate to
    :param alpha: how far we want to interpolate (0=start, 1=end)
    :return: float, result of the linear interpolation
    """

    # return (1 - alpha) * start + alpha * end
    return start + alpha * (end - start)


def clamp(number, min_value=0.0, max_value=1.0):
    """
    Clamps a number between two values
    :param number: number, value to clamp
    :param min_value: number, maximum value of the number
    :param max_value: number, minimum value of the number
    :return: variant, int || float
    """

    return max(min(number, max_value), min_value)


def remap_value(value, old_min, old_max, new_min, new_max):
    """
    Remap value taking into consideration its old range and the new one
    :param value: float
    :param old_min: float
    :param old_max: float
    :param new_min: float
    :param new_max: float
    :return: float
    """

    return new_min + (value - old_min) * (new_max - new_min) / (old_max - old_min)


def roundup(number, to):
    """
    Round up a number
    :param number: number to roundup
    :param to: number, mas value to roundup
    :return: variant, int || float
    """

    return int(math.ceil(number / to)) * to


def sign(value):
    """
    Returns the sign of the given value
    :param value: float
    :return: -1 of the value is negative; 1 if the value is positive; 0 if the value is zero
    """

    return value and (1, -1)[value < 0]


def get_range_percentage(min_value, max_value, value):
    """
    Returns the percentage value along a line from min_vlaue to max_value that value is
    :param min_value: float, minimum value
    :param max_value: float, maximum value
    :param value: float, input value
    :return: Percentage (from 0.0 to 1.0) between the two values where input value is
    """

    try:
        return (value - min_value) / (max_value - min_value)
    except ZeroDivisionError:
        return 0.0


def map_range_clamped(value, in_range_a, in_range_b, out_range_a, out_range_b):
    """
    Returns value mapped from one range into another where the value is clamped to the input range
    For example, 0.5 normalized from the range 0:1 to 0:50 would result in 25
    :param value: float
    :param in_range_a: float
    :param in_range_b: float
    :param out_range_a: float
    :param out_range_b: float
    :return: float
    """

    clamped_percentage = clamp(get_range_percentage(in_range_a, in_range_b, value), 0.0, 1.0)
    return lerp(out_range_a, out_range_b, clamped_percentage)


def map_range_unclamped(value, in_range_a, in_range_b, out_range_a, out_range_b):
    """
    Returns value mapped from one range into another where the value
    For example, 0.5 normalized from the range 0:1 to 0:50 would result in 25
    :param value: float
    :param in_range_a: float
    :param in_range_b: float
    :param out_range_a: float
    :param out_range_b: float
    :return: float
    """

    clamped_percentage = get_range_percentage(in_range_a, in_range_b, value)
    return lerp(out_range_a, out_range_b, clamped_percentage)


def bounding_box_half_values(bbox_min, bbox_max):
    """
    Returns the values half way between max and min XYZ given tuples
    :param bbox_min: tuple, contains the minimum X,Y,Z values of the mesh bounding box
    :param bbox_max: tuple, contains the maximum X,Y,Z values of the mesh bounding box
    :return: tuple(int, int, int)
    """

    min_x, min_y, min_z = bbox_min
    max_x, max_y, max_z = bbox_max
    half_x = (min_x + max_x) * 0.5
    half_y = (min_y + max_y) * 0.5
    half_z = (min_z + max_z) * 0.5

    return half_x, half_y, half_z


def snap_value(input, snap_value):
    """
    Returns snap value given an input and a base snap value
    :param input: float
    :param snap_value: float
    :return: float
    """

    return round((float(input) / snap_value)) * snap_value


def check_vector(vector):
    """
    Returns new Vector object from the given vector
    :param vector: variant, list<float, float, float> || Vector
    :return: Vector
    """

    if isinstance(vector, Vector):
        return vector

    return Vector(vector[0], vector[1], vector[2])


def check_vector_2d(vector):
    """
    Returns new Vector2D object from the given vector
    :param vector: variant, list<float, float> || Vector
    :return: Vector
    """

    if isinstance(vector, Vector2D):
        return vector

    return Vector(vector[0], vector[1])


def vector_add(vector1, vector2):
    """
    Adds one vector to another
    :param vector1: list(float, float, float)
    :param vector2: list(float, float, float)
    :return: list(float, float, float)
    """

    return [vector1[0] + vector2[0], vector1[1] + vector2[1], vector1[2] + vector2[2]]


def vector_sub(vector1, vector2):
    """
    Subtracts one vector to another
    :param vector1: list(float, float, float)
    :param vector2: list(float, float, float)
    :return: list(float, float, float)
    """

    return [vector1[0] - vector2[0], vector1[1] - vector2[1], vector1[2] - vector2[2]]


def vector_multiply(vector, value):
    """
    Multiples given vector by a value
    :param vector: list(float, float, float)
    :param value: float ,value to multiple vector by
    :return: list(float, float, float)
    """

    result = [vector[0] * value, vector[1] * value, vector[2] * value]

    return result


def vector_divide(vector, value):
    """
    Divides given vector by a value
    :param vector: list(float, float, float)
    :param value: float ,value to multiple vector by
    :return: list(float, float, float)
    """

    result = [vector[0] / value, vector[1] / value, vector[2] / value]

    return result


def vector_magnitude(vector):
    """
    Returns the magnitude of a vector
    :param vector: list(float, float, float)
    :return:  float
    """

    magnitude = math.sqrt(vector[0] ** 2 + vector[1] ** 2 + vector[2] ** 2)

    return magnitude


def vector_normalize(vector):
    """
    Normalizes given vector
    :param vector: list(float, float, float)
    :return: list(float, float, float)
    """

    return vector_divide(vector, vector_magnitude(vector))


def get_distance_between_vectors(vector1, vector2):
    """
    Returns the distance bewteen two vectors
    :param vector1: list(float, float, float)
    :param vector2: list(float, float, float)
    :return: float
    """

    vector1 = Vector(vector1)
    vector2 = Vector(vector2)
    vector = vector1 - vector2
    dst = vector()

    return math.sqrt((dst[0] * dst[0]) + (dst[1] * dst[1]) + (dst[2] * dst[2]))


def get_distance_between_vectors_before_sqrt(vector1, vector2):
    """
    Returns the distance bewteen two vectors before applying square root
    :param vector1: list(float, float, float)
    :param vector2: list(float, float, float)
    :return: float
    """

    vector1 = Vector(vector1)
    vector2 = Vector(vector2)
    vector = vector1 - vector2
    dst = vector()

    return (dst[0] * dst[0]) + (dst[1] * dst[1]) + (dst[2] * dst[2])


def get_distance_2d(vector1_2d, vector2_2d):
    """
    Returns the distance between two 2D vectors
    :param vector1_2d: Vector2D
    :param vector2_2d: Vector2D
    :return: float, distance between the two 2D vectors
    """

    v1 = check_vector_2d(vector1_2d)
    v2 = check_vector_2d(vector2_2d)

    v = v1 - v2
    dst = v()

    return math.sqrt(dst[0] * dst[0]) + (dst[1] * dst[1])


def get_dot_product(vector1, vector2):
    """
    Returns the dot product of the two vectors
    :param vector1: Vector
    :param vector2: Vector
    :return: float, dot product between the two vectors
    """

    v1 = check_vector(vector1)
    v2 = check_vector(vector2)
    return (v1.x * v2.x) + (v1.y * v2.y) + (v1.z * v2.z)


def get_dot_product_2d(vector1_2d, vector2_2d):
    """
    Returns the dot product of the two vectors
    :param vector1_2d: Vector2D
    :param vector2_2d: Vector2D
    :return: float, dot product between the two vectors
    """

    v1 = check_vector(vector1_2d)
    v2 = check_vector(vector2_2d)

    return (v1.x * v2.x) + (v1.y * v2.y)


def get_mid_point(vector1, vector2):
    """
    Get the mid vector between 2 vectors
    :param vector1: list<float, float, float>
    :param vector2: list<float, float, float>
    :return: list<float, float, float>, midpoint vector between vector1 and vector2
    """

    values = list()
    for i in range(0, 3):
        values.append(get_average([vector1[i], vector2[i]]))

    return values


def get_average(numbers):
    """
    Returns the average value of the given numbers list
    :param numbers: list<float>, list of the floats to get average from
    :return: float, average of the floats in numbers list
    """

    total = 0.0
    for num in numbers:
        total += num

    return total / len(numbers)


def get_inbetween_vector(vector1, vector2, percent=0.5):
    """
    Returns a vector inbetween vector1 and vector2 at the given percent
    :param vector1: list(float, float, float), vector
    :param vector2: list(float, float, float), vector
    :param percent: float, percent the vector should be between vector1 and vector2.
        - 0 percent will be exactly on vector1
        - 1 percent will be exactly on vector2
        - 0.5 percent will be exactly in the mid point between vector1 and vector2
    :return:  list(float, float, float), vector that represents the vector at the percentage between vector and vector2
    """

    vector1 = Vector(vector1)
    vector2 = Vector(vector2)
    percent = 1 - percent
    vector = ((vector1 - vector2) * percent) + vector2

    return vector()


def get_axis_vector(axis_name, offset=1):
    """
    Returns axis vector from its name
    :param axis_name: name of the axis ('X', 'Y' or 'Z')
    :param offset: float, offset to the axis, by default is 1
    :return: list (1, 0, 0) = X | (0, 1, 0) = Y | (0, 0, 1) = Z
    """

    if axis_name in ['X', 'x']:
        return offset, 0, 0
    elif axis_name in ['Y', 'y']:
        return 0, offset, 0
    elif axis_name in ['Z', 'z']:
        return 0, 0, 1


def fade_sine(percent_value):
    input_value = math.pi * percent_value

    return math.sin(input_value)


def fade_cosine(percent_value):
    percent_value = math.pi * percent_value

    return (1 - math.cos(percent_value)) * 0.5


def fade_smoothstep(percent_value):
    return percent_value * percent_value * (3 - 2 * percent_value)


def fade_sigmoid(percent_value):
    if percent_value == 0:
        return 0

    if percent_value == 1:
        return 1

    input_value = percent_value * 10 + 1

    return (2 / (1 + (math.e**(-0.70258 * input_value)))) - 1


def ease_in_sine(percent_value):
    return math.sin(1.5707963 * percent_value)


def ease_in_expo(percent_value):
    return (pow(2, 8 * percent_value) - 1) / 255


def ease_out_expo(percent_value, power=2):
    return 1 - pow(power, -8 * percent_value)


def ease_out_circ(percent_value):
    return math.sqrt(percent_value)


def ease_out_back(percent_value):
    return 1 + (--percent_value) * percent_value * (2.70158 * percent_value + 1.70158)


def ease_in_out_sine(percent_value):
    return 0.5 * (1 + math.sin(math.pi * (percent_value - 0.5)))


def easi_in_out_quart(percent_value):
    if percent_value < 0.5:
        percent_value *= percent_value
        return 8 * percent_value * percent_value
    else:
        percent_value -= 1
        percent_value *= percent_value
        return 1 - 8 * percent_value * percent_value


def ease_in_out_expo(percent_value):
    if percent_value < 0.5:
        return (math.pow(2, 16 * percent_value) - 1) / 510
    else:
        return 1 - 0.5 * math.pow(2, -16 * (percent_value - 0.5))


def ease_in_out_circ(percent_value):
    if percent_value < 0.5:
        return (1 - math.sqrt(1 - 2 * percent_value)) * 0.5
    else:
        return (1 + math.sqrt(2 * percent_value - 1)) * 0.5


def ease_in_out_back(percent_value):
    if percent_value < 0.5:
        return percent_value * percent_value * (7 * percent_value - 2.5) * 2
    else:
        return 1 + (percent_value - 1) * percent_value * 2 * (7 * percent_value + 2.5)


def average_position(pos1=(0.0, 0.0, 0.0), pos2=(0.0, 0.0, 0.0), weight=0.5):
    """
    Returns the average of the two given positions. You can weight between 0 (first input) or 1 (second_input)
    :param pos1: tuple, first input position
    :param pos2: tuple, second input position
    :param weight: float, amount to weight between the two input positions
    :return: tuple
    """

    return (
        pos1[0] + ((pos2[0] - pos1[0]) * weight),
        pos1[1] + ((pos2[1] - pos1[1]) * weight),
        pos1[2] + ((pos2[2] - pos1[2]) * weight)
    )


def smooth_step(value, range_start=0.0, range_end=1.0, smooth=1.0):
    """
    Interpolates between 2 float values using hermite interpolation
    :param value: float, value to smooth
    :param range_start: float, minimum value of interpolation range
    :param range_end: float, maximum value of interpolation range
    :param smooth: float, strength of the smooth applied to the value
    :return: float
    """

    # Get normalized value
    range_val = range_end - range_start
    normalized_val = value / range_val

    # Get smooth value
    smooth_val = pow(normalized_val, 2) * (3 - (normalized_val * 2))
    smooth_val = normalized_val + ((smooth_val - normalized_val) * smooth)
    value = range_start + (range_val * smooth_val)

    return value


def distribute_value(samples, spacing=1.0, range_start=0.0, range_end=1.0):
    """
    Returns a list of values distributed between a start and end range
    :param samples: int, number of values to sample across the value range
    :param spacing: float, incremental scale for each sample distance
    :param range_start: float, minimum value in the sample range
    :param range_end: float, maximum value in the sample range
    :return: list<float>
    """

    # Get value range
    value_list = [range_start]
    value_dst = abs(range_end - range_start)
    unit = 1.0

    # Find unit distance
    factor = 1.0
    for i in range(samples - 2):
        unit += factor * spacing
        factor *= spacing
    unit = value_dst / unit
    total_unit = unit

    # Build Sample list
    for i in range(samples - 2):
        mult_factor = total_unit / value_dst
        value_list.append(range_start - ((range_start - range_end) * mult_factor))
        unit *= spacing
        total_unit += unit

    # Append final sample
    value_list.append(range_end)

    return value_list


def inverse_distance_weight_1d(value_array, sample_value, value_domain=(0, 1), cycle_value=False):
    """
    Returns the inverse distance weight for a given sample point given an array of scalar values
    :param value_array: list<float>, value array to calculate weights from
    :param sample_value: float, sample point to calculate weights for
    :param value_domain: variant, tuple || list, minimum and maximum range of the value array
    :param cycle_value: bool, Whether to calculate or not the distance based on a closed loop of values
    :return: float
    """

    dst_array = list()
    total_inv_dst = 0.0

    # Calculate inverse distance weight
    for v in range(len(value_array)):
        dst = abs(sample_value - value_array[v])
        if cycle_value:
            value_domain_len = value_domain[1] - value_domain[0]
            f_cyc_dst = abs(sample_value - (value_array[v] + value_domain_len))
            r_cyc_dst = abs(sample_value - (value_array[v] - value_domain_len))
            if f_cyc_dst < dst:
                dst = f_cyc_dst
            if r_cyc_dst < dst:
                dst = r_cyc_dst

        # Check zero distance
        if dst < 0.00001:
            dst = 0.00001

        dst_array.append(dst)
        total_inv_dst += 1.0 / dst

    # Normalize value weights
    weight_array = [(1.0 / d) / total_inv_dst for d in dst_array]

    return weight_array


def rotation_matrix_xyz(rotation_angles):
    """
    Converts given rotation angles to a rotation represented by the sequences of rotations  around XYZ with given angles
    """

    rad_angles = [math.radians(x) for x in rotation_angles]
    x_angle = rad_angles[0]
    y_angle = rad_angles[1]
    z_angle = rad_angles[2]
    s1, c1 = math.sin(z_angle), math.cos(z_angle)
    s2, c2 = math.sin(y_angle), math.cos(y_angle)
    s3, c3 = math.sin(x_angle), math.cos(x_angle)

    m = ((c1 * c2, c1 * s2 * s3 - c3 * s1, s1 * s3 + c1 * c3 * s2),
         (c2 * s1, c1 * c3 + s1 * s2 * s3, c3 * s1 * s2 - c1 * s3),
         (- s2, c2 * s3, c2 * c3))

    return m


def rotation_matrix_xzy(rotation_angles):
    """
    Converts given rotation angles to a rotation represented by the sequences of rotations  around XZY with given angles
    """

    rad_angles = [math.radians(x) for x in rotation_angles]
    x_angle = rad_angles[0]
    y_angle = rad_angles[1]
    z_angle = rad_angles[2]
    s1, c1 = math.sin(z_angle), math.cos(z_angle)
    s2, c2 = math.sin(y_angle), math.cos(y_angle)
    s3, c3 = math.sin(x_angle), math.cos(x_angle)

    m = ((c1 * c2, s1 * s3 - c1 * c3 * s2, c3 * s1 + c1 * s2 * s3),
         (s2, c2 * c3, -c2 * s3),
         (-c2 * s1, c1 * s3 + c3 * s1 * s2, c1 * c3 - s1 * s2 * s3))

    return m


def rotation_matrix_to_xyz_euler(rotation_matrix):
    """
    Extracts XYZ euler angles from given rotation matrix
    """

    sy = math.sqrt(rotation_matrix[0][0] * rotation_matrix[0][0] + rotation_matrix[1][0] * rotation_matrix[1][0])
    singular = sy < 1e-7
    if not singular:
        x = math.degrees(math.atan2(rotation_matrix[2][1], rotation_matrix[2][2]))
        y = math.degrees(math.atan2(-rotation_matrix[2][0], sy))
        z = math.degrees(math.atan2(rotation_matrix[1][0], rotation_matrix[0][0]))
    else:
        x = math.degrees(math.atan2(-rotation_matrix[1][2], rotation_matrix[1][1]))
        y = math.degrees(math.atan2(-rotation_matrix[2][0], sy))
        z = 0

    return [x, y, z]


def rotation_matrix_to_xzy_euler(rotation_matrix):
    """
    Extracts XZY euler angles from given rotation matrix
    """

    sy = math.sqrt(rotation_matrix[0][0] * rotation_matrix[0][0] + rotation_matrix[2][0] * rotation_matrix[2][0])
    singular = sy < 1e-7
    if not singular:
        x = math.degrees(math.atan2(-rotation_matrix[1][2], rotation_matrix[1][1]))
        y = math.degrees(math.atan2(rotation_matrix[1][0], sy))
        z = math.degrees(math.atan2(-rotation_matrix[2][0], rotation_matrix[0][0]))
    else:
        x = math.degrees(math.atan2(-rotation_matrix[1][2], rotation_matrix[1][1]))
        y = math.degrees(math.atan2(rotation_matrix[1][0], sy))
        z = 0

    return [x, y, z]


def max_index(numbers):
    """
    Returns the largest number in the given list of numbers
    :param numbers: list(int) or list (float) or list(str)
    :return: int or float or str
    """

    max_value = 0
    result = 0
    for i in numbers:
        current_value = abs(float(i))
        if current_value > max_value:
            max_value = current_value
            result = numbers.index(i)

    return result



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utility methods related to write/read YAML files
"""


from __future__ import print_function, division, absolute_import

import os
import yaml
import logging

import yamlordereddictloader

LOGGER = logging.getLogger('tpDcc-libs-python')


def write_to_file(data, filename, **kwargs):

    """
    Writes data to JSON file
    """

    # if '.yml' not in filename:
    #     filename += '.yml'

    indent = kwargs.pop('indent', 2)

    try:
        with open(filename, 'w') as yaml_file:
            yaml.safe_dump(data, yaml_file, indent=indent, **kwargs)
    except IOError:
        LOGGER.error('Data not saved to file {}'.format(filename))
        return None

    LOGGER.debug('File correctly saved to: {}'.format(filename))

    return filename


def read_file(filename, maintain_order=False):

    """
    Get data from JSON file
    """

    if os.stat(filename).st_size == 0:
        return None
    else:
        try:
            with open(filename, 'r') as yaml_file:
                if maintain_order:
                    data = yaml.load(yaml_file, Loader=yamlordereddictloader.Loader)
                else:
                    data = yaml.safe_load(yaml_file)
        except Exception as err:
            LOGGER.warning('Could not read {0}'.format(filename))
            raise err

    return data



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains functions to handle Python modules
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import uuid
import pkgutil
import inspect
import logging
import importlib
import traceback

from tpDcc.libs.python import python, path as path_utils

if python.is_python3():
    from importlib.machinery import SourceFileLoader
else:
    import imp

LOGGER = logging.getLogger('tpDcc-libs-python')


def is_dotted_module_path(module_path):
    """
    Returns whether given module path is a dotted one (tpDcc.libs.python.modules) or not
    :param module_path: str
    :return: bool
    """

    return len(module_path.split('.')) > 2


def convert_to_dotted_path(path):
    """
    Returns a dotted path relative to the given path
    :param path: str, (eg. randomPath/folder/test.py)
    :return: str, dotted path (eg. folder.test)
    """

    directory, file_path = os.path.split(path)
    directory = path_utils.clean_path(directory)
    file_name = os.path.splitext(file_path)[0]
    package_path = [file_name]
    sys_path = list(set([path_utils.clean_path(p) for p in sys.path]))

    # We ignore current working directory. Useful if we want to execute tools directly inside PyCharm
    current_work_dir = path_utils.clean_path(os.getcwd())
    if current_work_dir in sys_path:
        sys_path.remove(current_work_dir)

    drive_letter = os.path.splitdrive(path)[0] + '\\'
    while directory not in sys_path:
        directory, name = os.path.split(directory)
        directory = path_utils.clean_path(directory)
        if directory == drive_letter or name == '':
            return ''
        package_path.append(name)

    return '.'.join(reversed(package_path))


def import_module(module_path, name=None, skip_warnings=False, skip_errors=False):
    """
    Imports the given module path. If the given module path is a dotted one, import lib will be used. Otherwise, it's
    expected that given module path is the absolute path to the source file. If name argument is not given, then the
    basename without the extension will be used
    :param module_path: str, module path. Can be a dotted path (tpDcc.libs.python.modules) or an absolute one
    :param name: str, name for the imported module which will be used if the module path is an absolute path
    :param skip_warnings: bool, Whether or not warnings should be skipped
    :param skip_errors: bool, Whether or not errors should be skipped
    :return: ModuleObject, imported module object
    """

    if is_dotted_module_path(module_path) and not os.path.exists(module_path):
        try:
            return importlib.import_module(module_path)
        except ImportError:
            if not skip_errors:
                LOGGER.error('Failed to load module: "{}"'.format(module_path), exc_info=True)
            return None

    try:
        if os.path.exists(module_path):
            if not name:
                name = os.path.splitext(os.path.basename(module_path))[0]
            if name in sys.modules:
                return sys.modules[name]
        if not name:
            if not skip_warnings:
                LOGGER.warning(
                    'Impossible to load module because module path: {} was not found!'.format(module_path))
            return None
        if os.path.isdir(module_path):
            module_path = os.path.join(module_path, '__init__.py')
            if not os.path.exists(module_path):
                raise ValueError('Cannot find module path: "{}"'.format(module_path))
        if python.is_python3():
            return SourceFileLoader(name, os.path.realpath(module_path)).load_module()
        else:
            if module_path.endswith('.py'):
                return imp.load_source(name, os.path.realpath(module_path))
            elif module_path.endswith('.pyc'):
                return imp.load_compiled(name, os.path.realpath(module_path))
    except ImportError:
        LOGGER.error('Failed to load module: "{}"'.format(module_path))
        raise


def resolve_module(name, log_error=False):
    """
    Resolves a dotted module name to a global object
    :param name: str
    :return:
    """

    name = name.split('.')
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used = used + '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            try:
                __import__(used)
            except ImportError:
                if log_error:
                    LOGGER.error(traceback.format_exc())
                return None
            found = getattr(found, n)

    return found


def iterate_modules(path, exclude=None, skip_inits=True, recursive=True, return_pyc=False):
    """
    Iterates all the modules of the given path
    :param path: str, folder path to iterate
    :param exclude: list(str), list of files to exclude
    :return: iterator
    """

    exclude = python.force_list(exclude)
    _exclude = ['__init__.py', '__init__.pyc'] if skip_inits else list()

    modules_found = dict()

    extension_to_skip = '.pyc' if not return_pyc else '.py'

    if recursive:
        for root, dirs, files in os.walk(path):
            if '__init__.py' not in files:
                continue
            for f in files:
                base_name = os.path.splitext(f)[0]
                if f not in exclude and base_name not in exclude:
                    module_path = path_utils.clean_path(os.path.join(root, f))
                    if f.endswith('.py') or f.endswith('.pyc') and base_name:
                        if base_name in modules_found:
                            if base_name.endswith(extension_to_skip):
                                continue
                            else:
                                if base_name.endswith(extension_to_skip) and base_name not in modules_found:
                                    modules_found[base_name] = module_path
                        else:
                            modules_found[base_name] = module_path
    else:
        files = os.listdir(path)
        if '__init__.py' not in files:
            return list(modules_found.values())
        for file_name in files:
            base_name = os.path.splitext(file_name)[0]
            if file_name not in exclude and base_name not in exclude:
                module_path = path_utils.clean_path(os.path.join(path, file_name))
                if file_name.endswith('.py') or file_name.endswith('.pyc') and base_name:
                    if base_name in modules_found:
                        if base_name.endswith(extension_to_skip):
                            continue
                        else:
                            if base_name.endswith(extension_to_skip) and base_name not in modules_found:
                                modules_found[base_name] = module_path
                    else:
                        modules_found[base_name] = module_path

    return list(modules_found.values())


def iterate_module_members(module_to_iterate, predicate=None):
    """
    Iterates all the members of the given modules
    :param module_to_iterate: ModuleObject, module object to iterate members of
    :param predicate: inspect.cass, if given members will be restricted to given inspect class
    :return: iterator
    """

    for mod in inspect.getmembers(module_to_iterate, predicate=predicate):
        yield mod


def iterate_module_subclasses(module, class_type):
    """
    Iterates all classes within a module object returning all subclasses of given type
    :param module: ModuleObject, module object to iterate subclasses of
    :param class_type: object, class object to find
    :return: generator(object), generator function returning class objects
    """

    for member in iterate_module_members(module, predicate=inspect.isclass):
        if issubclass(member, class_type):
            yield member


def get_package_children(module_path):
    """
    Returns all package children in given module path
    :param module_path: str
    :return: list(str)
    """

    return [name for _, name, _ in pkgutil.iter_modules([module_path])]


def load_module_from_source(file_path, unique_namespace=False):
    """
    Loads a Python from given source file
    :param file_path:
    :param unique_namespace: bool
    :return:
    """

    file_name = os.path.splitext(os.path.basename(file_path))[0]

    module_name = '{}{}'.format(file_name, str(uuid.uuid4())) if unique_namespace else file_name

    try:
        if python.is_python2():
            if file_path.endswith('.py'):
                return imp.load_source(module_name, file_path)
            elif file_path.endswith('.pyc'):
                return imp.load_compiled(module_name, file_path)
        else:
            return SourceFileLoader(module_name, file_path).load_module()
    except BaseException:
        LOGGER.debug('Failed trying to direct load : {} | {}'.format(file_path, sys.exc_info()))
        return None



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tests for tpDcc-libs-python
"""

import pytest

from tpDcc.libs.python import __version__


def test_version():
    assert __version__.get_version()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different functions related with security
"""

from __future__ import print_function, division, absolute_import


import base64


def encodeBase64(key, clear):
    enc = list()
    for i in range(len(clear)):
        key_c = key[i % len(key)]
        enc_c = chr((ord(clear[i]) + ord(key_c)) % 256)
        enc.append(enc_c)

    return base64.urlsafe_b64encode(''.join(enc))


def decodeBase64(key, enc):
    dec = list()
    enc = base64.urlsafe_b64decode(enc)
    for i in range(len(enc)):
        key_c = key[i % len(key)]
        dec_c = chr((256 + ord(enc[i]) - ord(key_c)) % 256)
        dec.append(dec_c)

    return ''.join(dec)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementations for Plugin Factory mechanism
"""

from __future__ import print_function, division, absolute_import

import os
import re
import sys
import inspect
import logging
from distutils import version

from tpDcc.libs.python import python, modules, path as path_utils, folder as folder_utils

LOGGER = logging.getLogger('tpDcc-libs-python')


class PluginFactory(object):

    class PluginLoadingMechanism(object):
        """
        Class that contains variables used to define how the plugins on a registered path should be loaded
        """

        # Default mechanism. It will attempt to use the IMPORTABLE method first, and if the module is not
        # accessible from within sys modules it will fall back to LOAD_SOURCE.
        GUESS = 0

        # Mechanism to use when your plugin code is outside of the interpreter sys path. The plugin file
        # be loaded instead of imported. Gives flexibility in terms of structure but you cannot use
        # relative import paths within the plugins loaded with this mechanism. All loaded plugins using
        # this module are imported into a namespace defined through a uuid.
        LOAD_SOURCE = 1

        # Mechanism to use when your plugin code resides within already importable locations. Mandatory
        # to use when your plugin contains relative imports. Because this is importing modules which
        # are available on the sys.path, the class names will resolve nicely too
        IMPORTABLE = 2

    # Regex validator for plugin folder directories
    REGEX_FOLDER_VALIDATOR = re.compile('^((?!__pycache__).)*$')

    # Regex validator for plugin file names
    REGEX_FILE_VALIDATOR = re.compile('([a-zA-Z].*)(\.py$|\.pyc$)')

    def __init__(self, interface, paths=None, package_name=None, plugin_id=None, version_id=None, env_var=None):
        """

        :param interface: Abstract class to use when searching for plugins within the registered paths.
        :param paths: list(str), list of absolute paths to search for plugins.
        :param plugin_id: str, plugin identifier to distinguish between different plugins. If not given, plugin
            class name will be used
        :param version_id: str, plugin version identifier. If given, allows plugins with the same identifier to be
            differentiated.
        :param env_var: str, optional environment variable name containing paths to register separated by OS separator.
        """

        self._interface = interface
        self._plugin_identifier = plugin_id or '__name__'
        self._version_identifier = version_id

        self._plugins = dict()
        self._registered_paths = dict()

        self.register_paths(paths, package_name=package_name)
        if env_var:
            self.register_paths_from_env_var(env_var, package_name=package_name)

    def __repr__(self):
        return '[{} - Identifier: {}, Plugin Count: {}]'.format(
            self.__class__.__name__, self._plugin_identifier, len(self._plugins))

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def register_path(self, path, package_name=None, mechanism=PluginLoadingMechanism.GUESS):
        """
        Registers a search path within the factory. The factory will immediately being searching recursively withing
        this location for any plugin.
        :param path: str, absolute path to register into the factory
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        :param mechanism: PluginLoadingMechanism, plugin load mechanism to use
        :return: int, total amount of registered plugins
        """

        if not path or not os.path.isdir(path):
            return 0

        package_name = package_name or 'tpDcc'

        # Regardless of what is found in the given path, we store it
        self._registered_paths.setdefault(package_name, dict()).setdefault(path, dict())
        self._registered_paths[package_name][path] = mechanism

        current_plugins_count = len(self._plugins)

        file_paths = list()
        for root, _, files in folder_utils.walk_level(path, level=1):
            if not self.REGEX_FOLDER_VALIDATOR.match(root):
                continue

            for file_name in files:

                # Skip files that do not match PluginFactory regex validator.
                if not self.REGEX_FILE_VALIDATOR.match(file_name):
                    continue

                file_paths.append(path_utils.clean_path(os.path.join(root, file_name)))

        # Loop through all the found files searching for plugins definitions
        for file_path in file_paths:

            module_to_inspect = None

            if mechanism in (self.PluginLoadingMechanism.IMPORTABLE, self.PluginLoadingMechanism.GUESS):
                module_to_inspect = self._mechanism_import(file_path)
                if module_to_inspect:
                    LOGGER.debug('Module Import : {}'.format(file_path))

            if not module_to_inspect:
                if mechanism in (self.PluginLoadingMechanism.LOAD_SOURCE, self.PluginLoadingMechanism.GUESS):
                    module_to_inspect = self._mechanism_load(file_path)
                    if module_to_inspect:
                        LOGGER.debug('Direct Load : {}'.format(file_path))

            if not module_to_inspect:
                LOGGER.debug('Could not import or load : {}'.format(file_path))
                continue

            try:
                for item_name in dir(module_to_inspect):

                    item = getattr(module_to_inspect, item_name)
                    if inspect.isclass(item):
                        if item == self._interface:
                            continue
                        if issubclass(item, self._interface):
                            item.ROOT = path
                            item.PATH = file_path
                            self._plugins.setdefault(package_name, list())
                            self._plugins[package_name].append(item)
            except BaseException:
                LOGGER.debug('', exc_info=True)

        return len(self._plugins) - current_plugins_count

    def register_paths(self, paths, package_name=None, mechanism=PluginLoadingMechanism.GUESS):
        """
        Registers given paths within the factory. The factory will immediately being searching recursively withing
        this location for any plugin.
        :param paths: list(str), absolute paths to register into the factory
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        :param mechanism: PluginLoadingMechanism, plugin load mechanism to use
        :return: int, total amount of registered plugins
        """

        paths = python.force_list(paths)

        total_plugins = 0
        visited = set()
        for path in paths:
            if not path:
                continue
            base_name = path_utils.clean_path(os.path.splitext(path)[0] if os.path.isfile(path) else path)
            if base_name in visited:
                continue
            visited.add(base_name)
            total_plugins += self.register_path(path, package_name=package_name, mechanism=mechanism)

        return total_plugins

    def register_paths_from_env_var(self, env_var, package_name=None, mechanism=PluginLoadingMechanism.GUESS):
        """
        Registers paths contained in given environment variables. Paths must be separated with OS separator
        :param env_var: str, environment variable we are going to retrieve paths from
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        :param mechanism: PluginLoadingMechanism, plugin load mechanism to use
        :return: int, total amount of registered plugins
        """

        paths = os.environ.get(env_var, '').split(os.pathsep)
        if not paths:
            return

        return self.register_paths(paths, package_name=package_name, mechanism=mechanism)

    def register_plugin_from_class(self, plugin_class, package_name=None):
        """
        Registers the given class type as a plugin for this factory. Note that the given type class must be inherited
        from the factory interface. Useful when you have direct access to the plugin classes without the need of
        searching disk locations (which is slow)
        :param plugin_class: type, class type to add to the factory
        :return: True if the registration is successful; False otherwise.
        """

        if not inspect.isclass(plugin_class) or not issubclass(plugin_class, self._interface):
            return False

        if not package_name:
            class_id = self._get_identifier(plugin_class)
            split_id = class_id.replace('.', '-').split('-')[0]
            package_name = split_id if split_id != class_id else 'tpDcc'

        self._plugins.setdefault(package_name, list).append(plugin_class)

        return True

    def paths(self, package_name=None):
        """
        Returns all registered paths in the factory
        :return: list(str)
        """

        return list(self._registered_paths.get(package_name, dict()).keys())

    def identifiers(self, package_name=None):
        """
        Returns a list of plugin class names within the factory. The list of class names is unique, so classes which
        share the same name will not appear twice.
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        :return: list(str)
        """

        package_name = package_name or 'tpDcc'
        return {
            self._get_identifier(plugin) for plugin in self._plugins.get(package_name, list())
        }

    def versions(self, identifier, package_name=None):
        """
        Returns a list of all the versions available for the plugins with the given identifier
        :param identifier: str, Plugin identifier to check
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        :return: list(str)
        """

        if not self._version_identifier:
            return list()

        return sorted(
            self._get_version(plugin) for plugin in self._plugins.get(
                package_name, list()) if self._get_identifier(plugin) == identifier
        )

    def plugins(self, package_name=None):
        """
        Returns a unique list of plugins. Where multiple versions are available the highest version will be given
        :param package_name: str, package name current registered plugins will belong to.
        :return: list(class)
        """

        return [self.get_plugin_from_id(
            identifier, package_name=package_name) for identifier in self.identifiers(package_name=package_name)]

    def get_plugin_from_id(self, plugin_id, package_name=None, plugin_version=None):
        """
        Retrieves the plugin with given plugin identifier. If you require a specific version of a plugin (in a
        scenario where there are multiple plugins with the same identifier) this can also be specified
        :param plugin_id: str, identifying value of the plugin you want to retrieve
        :param package_name: str, package name current registered plugins will belong to.
        :param plugin_version: int or float, version of the plugin you want. If factory has no versioning identifier
            specified this argument has no effect.
        :return: Plugin
        """

        if package_name and package_name not in self._plugins:
            LOGGER.error('Impossible to retrieve data from id: {} package: "{}" not registered!'.format(
                plugin_id, package_name))
            return None

        if package_name:
            matching_plugins = [plugin for plugin in self._plugins.get(
                package_name, list()) if self._get_identifier(plugin) == plugin_id]
        else:
            matching_plugins = list()
            for plugins in list(self._plugins.values()):
                for plugin in plugins:
                    if self._get_identifier(plugin) == plugin_id:
                        matching_plugins.append(plugin)

        if not matching_plugins:
            LOGGER.warning('No plugin with id "{}" found in package "{}"'.format(plugin_id, package_name))
            return None

        if not self._version_identifier:
            return matching_plugins[0]

        versions = {
            self._get_version(plugin): plugin for plugin in matching_plugins
        }
        ordered_versions = [version.LooseVersion(str(v)) for v in list(versions.keys())]

        # If not version given, we return the plugin with the highest value
        if not plugin_version:
            return versions[str(ordered_versions[0])]

        plugin_version = version.LooseVersion(plugin_version)
        if plugin_version not in ordered_versions:
            LOGGER.warning('No Plugin with id "{}" and version "{}" found in package "{}"'.format(
                plugin_id, plugin_version, package_name))
            return None

        return versions[str(plugin_version)]

    def unregister_path(self, path, package_name=None):
        """
        Unregister given path from the list of registered paths
        :param path: Absolute path we want to unregister from registered factory paths
        :param package_name: str, package name current registered plugins will belong to. Default to tDcc.
        """

        registered_paths = self._registered_paths.copy()

        self._plugins = list()
        self._registered_paths = dict()

        for pkg_name, registered_paths_dict in registered_paths.items():
            for original_path, mechanism in registered_paths_dict.items():
                if package_name == pkg_name and path_utils.clean_path(original_path) == path_utils.clean_path(path):
                    continue
                self.register_path(original_path, package_name=pkg_name, mechanism=mechanism)

    def reload(self):
        """
        Clears all registered plugins and performs a search over all registered paths
        """

        registered_paths = self._registered_paths.copy()

        self.clear()

        for package_name, registered_paths_dict in registered_paths.items():
            for original_path, mechanism in registered_paths_dict.items():
                self.register_path(original_path, package_name=package_name, mechanism=mechanism)

    def clear(self):
        """
        Clears all the plugins and registered paths
        """

        self._plugins.clear()
        self._registered_paths.clear()

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _mechanism_import(self, file_path):
        """
        Internal function that will try to retrieve a module from a given path by looking current sys.path
        :param file_path: str, absolute file path of a Python file
        :return: module or None
        """

        # In Python 2 we check the existence of an __init__ file
        has_init = True
        if python.is_python2():
            has_init = False
            file_dir = os.path.dirname(file_path)
            for extension in ('.py', '.pyc'):
                if os.path.isfile(os.path.join(file_dir, '__init__{}'.format(extension))):
                    has_init = True
                    break
        if not has_init:
            return None

        module_name = modules.convert_to_dotted_path(file_path)
        if module_name:
            try:
                return sys.modules[module_name]
            except KeyError:
                return modules.import_module(module_name, skip_errors=True)

        return None

    def _mechanism_load(self, file_path):
        """
        Internal function that will try to retrieve a module by directly loading its source code
        :param file_path: str, absolute file path of a Python file
        :return: module or None
        """

        return modules.load_module_from_source(file_path)

    def _get_identifier(self, plugin):
        """
        Internal function that uses plugin identifier to request the identifying name of the plugin
        :param plugin: str, plugin to take name from
        :return: str
        """

        identifier = getattr(plugin, self._plugin_identifier)

        predicate = inspect.ismethod if python.is_python2() else inspect.isfunction
        if predicate(identifier):
            return identifier()

        return identifier

    def _get_version(self, plugin):
        """
        Internal function that uses plugin version identifier to request the version number of the plugin
        :param plugin: str, plugin to take version from
        :return: int or float
        """

        identifier = getattr(plugin, self._version_identifier)

        predicate = inspect.ismethod if python.is_python2() else inspect.isfunction
        if predicate(identifier):
            return str(identifier())

        return str(identifier)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that functions to retrieve info from running processes
"""

from __future__ import print_function, division, absolute_import

import logging

PSUTIL_AVAILABLE = True
try:
    import psutil
except Exception:
    PSUTIL_AVAILABLE = False

logger = logging.getLogger('tpDcc-libs-python')


def check_if_process_is_running(process_name):
    """
    Returns whether or not a process with given name is running
    :param process_name: str
    :return: bool
    """

    if not PSUTIL_AVAILABLE:
        logger.warning(
            'Impossible to check is process "{}" is running because psutil is not available!'.format(process_name))
        return False

    for process in psutil.process_iter():
        try:
            if process_name.lower() in process.name().lower():
                return True
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains observer pattern related functions and classes
"""

from __future__ import print_function, division, absolute_import

import time
import logging

LOGGER = logging.getLogger('tpDcc-libs-python')


class Observable(object):
    """
    Basic observable/observer pattern implementation
    """

    def __init__(self):

        # Dictionary that maps instance references with dictionaries where the key is the event id and its value
        # it is the callback that should be called
        self._connected = dict()

    def connect(self, instance, callback, event_id):
        """
        Connects instance to an event
        :param instance: object
        :param callback: function, function that will be called when teh event happens
        :param event_id: id, unique identifier for the event
        """

        self._connected.setdefault(instance, dict())
        self._connected[instance][event_id] = callback

    def disconnect(self, instance, event_id=None):
        """
        Disconnects instance from an event
        :param instance: object
        :param event_id: id, unique identifier of the event
        """

        if instance not in self._connected:
            return
        if event_id is None:
            self._connected[instance] = dict()
        else:
            del self._connected[instance][event_id]

    def emit(self, event_id, arg=None):
        """
        Emits event and all connected instances will be notified
        :param event_id: int
        :param arg:
        """

        if not self._connected:
            return
        start = time.perf_counter()
        for k in self._connected:
            if event_id in self._connected[k]:
                self._connected[k][event_id](arg)
        stop = time.perf_counter()
        if (stop - start) > 0.100:
            LOGGER.warning('*** SLOW *** {}.broadcast(event_id={}'.format(self.__class__.__name__, event_id) + (
                ',{}'.format(arg) if arg else '') + ') took %.2f' % (stop - start))
